#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include "mockturtle/libmockturtle.h"
#include "libwrtd.h"
#include "libwrtd-private.h"

/* Number of existing alarms.  */
enum wrtd_status wrtd_attr_get_alarm_count(struct wrtd_dev *wrtd,
                                           int32_t *value)
{
        enum wrtd_status status;
        unsigned i;
        int32_t res;

        status = wrtd_fill_alarms(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        res = 0;
        for (i = 0; i < wrtd->nbr_alarms; i++)
                if (!wrtd_id_null(wrtd->alarms[i].alarm.event.id))
                        res++;
        *value = res;
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_alarm_check_disabled(struct wrtd_dev *wrtd, unsigned idx)
{
        if (wrtd->alarms[idx].alarm.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RESOURCE_ACTIVE,
                                         "alarm is enabled");
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_rule_check_disabled(struct wrtd_dev *wrtd, unsigned idx)
{
        if (wrtd->rules[idx].rule.conf.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RESOURCE_ACTIVE,
                                         "rule is enabled");
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_period(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value->seconds != 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "alarm period is too large (>1s)");

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.period_ns = value->ns;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_alarm_period(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, wrtd->alarms[idx].alarm.period_ns);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_time(struct wrtd_dev *wrtd,
                                          const char *rep_cap_id,
                                          const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.event.ts = *value;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_alarm_time(struct wrtd_dev *wrtd,
                                          const char *rep_cap_id,
                                          struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->alarms[idx].alarm.event.ts;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_enable(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            bool value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.enabled = value;

        /* FIXME: may modify the rule mapping (eg: alarm is placed on cpu,
           and need to trigger an rule initially placed on a different cpu).
        */
        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_alarm_enable(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            bool *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->alarms[idx].alarm.enabled != 0;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_repeat_count(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  int32_t value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value < 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "alarm repeat count < 0");

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.repeat_count = value == 0 ? 0 : value + 1;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_alarm_repeat_count(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->alarms[idx].alarm.repeat_count;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_repeat_count(struct wrtd_dev *wrtd,
                                                 const char *rep_cap_id,
                                                 int32_t value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value < 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "rule repeat count < 0");

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->rules[idx].rule.conf.repeat_count = value == 0 ? 0 : value + 1;

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_repeat_count(struct wrtd_dev *wrtd,
                                                 const char *rep_cap_id,
                                                 int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.conf.repeat_count;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_log_enable(struct wrtd_dev *wrtd,
                                          bool value)
{
        enum wrtd_status status;
        unsigned int cpu;

        status = wrtd_fill_roots(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                wrtd->roots[cpu].log_flags = value ? 0xff : 0;
                status = wrtd_write_root_flags(wrtd, cpu);
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_log_enable(struct wrtd_dev *wrtd,
                                          bool *value)
{
        enum wrtd_status status;

        status = wrtd_fill_roots(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* all cpus have the same flags, it's enough to read one */
        *value = wrtd->roots[0].log_flags != 0;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_log_empty(struct wrtd_dev *wrtd,
                                         bool *value)
{
        struct polltrtl p[WRTD_MAX_CPUS];
        int ret, i;

        for (i = 0; i < wrtd->nbr_cpus; ++i) {
                p[i].trtl = wrtd->trtl;
                p[i].idx_hmq = WRTD_HMQ;
                p[i].idx_cpu = i;
                p[i].events = POLLIN;
        }

        ret = trtl_msg_poll(p, wrtd->nbr_cpus, 0);
        if (ret < 0)
          return WRTD_ERROR_UNEXPECTED_RESPONSE;

        *value = (ret == 0);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_time_sync(struct wrtd_dev *wrtd,
                                         bool *value)
{
        struct wrtd_config_msg msg;
        enum wrtd_status status;

        status = wrtd_msg_get_config(wrtd, 0, &msg);
        WRTD_RETURN_IF_ERROR(status);

        *value = (msg.sync_flag != 0);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_find_rule_check(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      unsigned *idx)
{
        enum wrtd_status status;

        status = wrtd_find_rule(wrtd, rep_cap_id, idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, *idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_count(struct wrtd_dev *wrtd,
                                          int32_t *value)
{
        enum wrtd_status status;
        unsigned i;
        int32_t res;

        status = wrtd_fill_rules(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        res = 0;
        for (i = 0; i < wrtd->nbr_rules; i++)
                if (!wrtd_id_null(wrtd->rules[i].rule.conf.id))
                        res++;
        *value = res;
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_source(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           const char *src)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_validate_id(wrtd, src);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_find_rule_check(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd_id_copy(wrtd->rules[idx].rule.conf.source_id, src);

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_destination(struct wrtd_dev *wrtd,
                                                const char *rep_cap_id,
                                                const char *dst)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_validate_id(wrtd, dst);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_find_rule_check(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd_id_copy(wrtd->rules[idx].rule.conf.dest_id, dst);

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_destination(struct wrtd_dev *wrtd,
                                                const char *rep_cap_id,
                                                int32_t buf_size,
                                                char *dst)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_id_copy_buf(wrtd, dst, buf_size,
                                  wrtd->rules[idx].rule.conf.dest_id);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_source(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           int32_t buf_size,
                                           char *dst)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_id_copy_buf(wrtd, dst, buf_size,
                                  wrtd->rules[idx].rule.conf.source_id);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_delay(struct wrtd_dev *wrtd,
                                          const char *rep_cap_id,
                                          const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value->seconds != 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "rule delay is too large (>1s)");

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->rules[idx].rule.conf.delay_ns = value->ns;

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_delay(struct wrtd_dev *wrtd,
                                          const char *rep_cap_id,
                                          struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, wrtd->rules[idx].rule.conf.delay_ns);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_holdoff(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value->seconds != 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "rule holdoff is too large (>1s)");

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->rules[idx].rule.conf.hold_off_ns = value->ns;

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_holdoff(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, wrtd->rules[idx].rule.conf.hold_off_ns);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_resync_period(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value->seconds != 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "rule resync period is too large (>1s)");

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->rules[idx].rule.conf.resync_period_ns = value->ns;

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_resync_period(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, wrtd->rules[idx].rule.conf.resync_period_ns);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_rx_last(struct wrtd_dev *wrtd,
                                                 const char *rep_cap_id,
                                                 struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.rx_last;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_tx_last(struct wrtd_dev *wrtd,
                                                 const char *rep_cap_id,
                                                 struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.tx_last;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_missed_last(struct wrtd_dev *wrtd,
                                                     const char *rep_cap_id,
                                                     struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.miss_last;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_rx_latency_min(struct wrtd_dev *wrtd,
                                                        const char *rep_cap_id,
                                                        struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, wrtd->rules[idx].rule.stat.lat_min_ns);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_rx_latency_max(struct wrtd_dev *wrtd,
                                                        const char *rep_cap_id,
                                                        struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, wrtd->rules[idx].rule.stat.lat_max_ns);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_rx_latency_avg(struct wrtd_dev *wrtd,
                                                        const char *rep_cap_id,
                                                        struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;
        uint64_t latency;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        latency   = wrtd->rules[idx].rule.stat.lat_hi_ns;
        latency <<= 32;
        latency  |= wrtd->rules[idx].rule.stat.lat_lo_ns;
        latency  /= wrtd->rules[idx].rule.stat.lat_nbr;

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, latency & 0xffffffff);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_rx_period_avg(struct wrtd_dev *wrtd,
                                                       const char *rep_cap_id,
                                                       struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;
        unsigned int period;
        struct wrtd_tstamp last;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        last = wrtd->rules[idx].rule.stat.rx_last;

        period  = (last.seconds * 1000000000 + last.ns);
        period /= wrtd->rules[idx].rule.stat.rx_events;

        memset(value, 0, sizeof(struct wrtd_tstamp));
        wrtd_ts_add_ns(value, period);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_resync_factor(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  int32_t value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->rules[idx].rule.conf.resync_factor = value;

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_resync_factor(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.conf.resync_factor;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_rx_events(struct wrtd_dev *wrtd,
                                                   const char *rep_cap_id,
                                                   int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.rx_events;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_tx_events(struct wrtd_dev *wrtd,
                                                   const char *rep_cap_id,
                                                   int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.tx_events;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_missed_late(struct wrtd_dev *wrtd,
                                                     const char *rep_cap_id,
                                                     int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.miss_late;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_missed_holdoff(struct wrtd_dev *wrtd,
                                                        const char *rep_cap_id,
                                                        int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.miss_holdoff;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_missed_nosync(struct wrtd_dev *wrtd,
                                                       const char *rep_cap_id,
                                                       int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.miss_nosync;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_stat_rule_missed_overflow(struct wrtd_dev *wrtd,
                                                         const char *rep_cap_id,
                                                         int32_t *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.stat.miss_overflow;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_enable(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           bool value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        if (wrtd->rules[idx].rule.conf.enabled != value) {
                wrtd->rules[idx].rule.conf.enabled = value;

                status = wrtd_reconfigure(wrtd);
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_enable(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           bool *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.conf.enabled != 0;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_send_late(struct wrtd_dev *wrtd,
                                              const char *rep_cap_id,
                                              bool value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        if (wrtd->rules[idx].rule.conf.send_late != value) {
                wrtd->rules[idx].rule.conf.send_late = value;

                status = wrtd_write_rule(wrtd, idx);
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_send_late(struct wrtd_dev *wrtd,
                                              const char *rep_cap_id,
                                              bool *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.conf.send_late != 0;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_global(struct wrtd_dev *wrtd,
                                  const char *rep_cap_id)
{
        if (!wrtd_id_eq(rep_cap_id, WRTD_GLOBAL_REP_CAP_ID))
                return wrtd_return_error(wrtd, WRTD_ERROR_ATTR_INVALID_REP_CAP,
                                         "global repeated capability required");
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_get_attr_int32(struct wrtd_dev *wrtd,
                                     const char *rep_cap_id,
                                     enum wrtd_attr id,
                                     int32_t *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_MAJOR_VERSION:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                *value = WRTD_VERSION_MAJOR;
                return WRTD_SUCCESS;
        case WRTD_MINOR_VERSION:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                *value = WRTD_VERSION_MINOR;
                return WRTD_SUCCESS;
        case WRTD_ATTR_ALARM_COUNT:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                status = wrtd_attr_get_alarm_count
                  (wrtd, value);
                WRTD_RETURN_IF_ERROR(status);
                return WRTD_SUCCESS;
        case WRTD_ATTR_ALARM_REPEAT_COUNT:
                return wrtd_attr_get_alarm_repeat_count
                        (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_COUNT:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                status = wrtd_attr_get_rule_count
                  (wrtd, value);
                WRTD_RETURN_IF_ERROR(status);
                return WRTD_SUCCESS;
        case WRTD_ATTR_RULE_REPEAT_COUNT:
                return wrtd_attr_get_rule_repeat_count
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_RESYNC_FACTOR:
                return wrtd_attr_get_rule_resync_factor
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_EVENTS:
                return wrtd_attr_get_stat_rule_rx_events
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_TX_EVENTS:
                return wrtd_attr_get_stat_rule_tx_events
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_MISSED_LATE:
                return wrtd_attr_get_stat_rule_missed_late
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_MISSED_HOLDOFF:
                return wrtd_attr_get_stat_rule_missed_holdoff
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_MISSED_NOSYNC:
                return wrtd_attr_get_stat_rule_missed_holdoff
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_MISSED_OVERFLOW:
                return wrtd_attr_get_stat_rule_missed_overflow
                  (wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "int32 attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_tstamp(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      const struct wrtd_tstamp *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_ALARM_TIME:
                return wrtd_attr_set_alarm_time
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_ALARM_PERIOD:
                return wrtd_attr_set_alarm_period
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_DELAY:
                return wrtd_attr_set_rule_delay
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_HOLDOFF:
                return wrtd_attr_set_rule_holdoff
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_RESYNC_PERIOD:
                return wrtd_attr_set_rule_resync_period
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_LAST:
        case WRTD_ATTR_STAT_RULE_TX_LAST:
        case WRTD_ATTR_STAT_RULE_MISSED_LAST:
        case WRTD_ATTR_STAT_RULE_RX_LATENCY_MIN:
        case WRTD_ATTR_STAT_RULE_RX_LATENCY_MAX:
        case WRTD_ATTR_STAT_RULE_RX_LATENCY_AVG:
        case WRTD_ATTR_STAT_RULE_RX_PERIOD_AVG:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_ATTR_NOT_WRITEABLE,
                         "attribute %u is not writeable", id);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "tstamp attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_get_attr_tstamp(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      struct wrtd_tstamp *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_ALARM_TIME:
                return wrtd_attr_get_alarm_time
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_ALARM_PERIOD:
                return wrtd_attr_get_alarm_period
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_DELAY:
                return wrtd_attr_get_rule_delay
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_HOLDOFF:
                return wrtd_attr_get_rule_holdoff
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_RESYNC_PERIOD:
                return wrtd_attr_get_rule_resync_period
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_LAST:
                return wrtd_attr_get_stat_rule_rx_last
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_TX_LAST:
                return wrtd_attr_get_stat_rule_tx_last
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_MISSED_LAST:
                return wrtd_attr_get_stat_rule_missed_last
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_LATENCY_MIN:
                return wrtd_attr_get_stat_rule_rx_latency_min
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_LATENCY_MAX:
                return wrtd_attr_get_stat_rule_rx_latency_max
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_LATENCY_AVG:
                return wrtd_attr_get_stat_rule_rx_latency_avg
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_STAT_RULE_RX_PERIOD_AVG:
                return wrtd_attr_get_stat_rule_rx_period_avg
                  (wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "tstamp attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_bool(struct wrtd_dev *wrtd,
                                    const char *rep_cap_id,
                                    enum wrtd_attr id,
                                    bool value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_EVENT_LOG_ENABLED:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                return wrtd_attr_set_log_enable
                  (wrtd, value);
        case WRTD_ATTR_ALARM_ENABLED:
                return wrtd_attr_set_alarm_enable
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_ENABLED:
                return wrtd_attr_set_rule_enable
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_SEND_LATE:
                return wrtd_attr_set_rule_send_late
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_EVENT_LOG_EMPTY:
        case WRTD_ATTR_IS_TIME_SYNCHRONIZED:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_ATTR_NOT_WRITEABLE,
                         "attribute %u is not writeable", id);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "bool attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_get_attr_bool(struct wrtd_dev *wrtd,
                                    const char *rep_cap_id,
                                    enum wrtd_attr id,
                                    bool *value)
{
       enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_EVENT_LOG_ENABLED:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                return wrtd_attr_get_log_enable
                  (wrtd, value);
        case WRTD_ATTR_EVENT_LOG_EMPTY:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                return wrtd_attr_get_log_empty
                  (wrtd, value);
        case WRTD_ATTR_IS_TIME_SYNCHRONIZED:
                status = wrtd_attr_global
                  (wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                return wrtd_attr_get_time_sync
                  (wrtd, value);
        case WRTD_ATTR_ALARM_ENABLED:
                return wrtd_attr_get_alarm_enable
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_ENABLED:
                return wrtd_attr_get_rule_enable
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_SEND_LATE:
                return wrtd_attr_get_rule_send_late
                  (wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "bool attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_int32(struct wrtd_dev *wrtd,
                                     const char *rep_cap_id,
                                     enum wrtd_attr id,
                                     int32_t value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_ALARM_REPEAT_COUNT:
                return wrtd_attr_set_alarm_repeat_count
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_REPEAT_COUNT:
                return wrtd_attr_set_rule_repeat_count
                  (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_RESYNC_FACTOR:
                return wrtd_attr_set_rule_resync_factor
                  (wrtd, rep_cap_id, value);
        case WRTD_MAJOR_VERSION:
        case WRTD_MINOR_VERSION:
        case WRTD_ATTR_ALARM_COUNT:
        case WRTD_ATTR_RULE_COUNT:
        case WRTD_ATTR_STAT_RULE_RX_EVENTS:
        case WRTD_ATTR_STAT_RULE_TX_EVENTS:
        case WRTD_ATTR_STAT_RULE_MISSED_LATE:
        case WRTD_ATTR_STAT_RULE_MISSED_HOLDOFF:
        case WRTD_ATTR_STAT_RULE_MISSED_NOSYNC:
        case WRTD_ATTR_STAT_RULE_MISSED_OVERFLOW:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_ATTR_NOT_WRITEABLE,
                         "attribute %u is not writeable", id);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "int32 attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_string(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      const char *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_RULE_SOURCE:
                return wrtd_attr_set_rule_source
                        (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_DESTINATION:
                return wrtd_attr_set_rule_destination
                        (wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "string set attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_get_attr_string(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      int32_t value_buf_size,
                                      char *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_RULE_SOURCE:
                return wrtd_attr_get_rule_source
                        (wrtd, rep_cap_id, value_buf_size, value);
        case WRTD_ATTR_RULE_DESTINATION:
                return wrtd_attr_get_rule_destination
                        (wrtd, rep_cap_id, value_buf_size, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "string get attribute %u is not known", id);
        }
}
